;; PittSFIeld ACL2 safety proof, version of February 13th, 2006
;; Stephen McCamant <smcc@csail.mit.edu>

(in-package "ACL2-PITTSFIELD")

;; First, we construct the model of the processor and the sandboxing
;; checks.

(in-theory (disable loghead))

(defun make-mem () nil)
(defun mem-fetch (mem addr)
  (if (bound? addr mem)
      (binding addr mem)
    #x00))
(defun mem-store (mem addr val)
  (bind addr val mem))

(defun x86-state-fetch (x86 addr)
  (mem-fetch (x86-state-mem x86) addr))

;; In the real system, we have the code region run from 0x10000000 to
;; 0x10ffffff (16MB). To allow faster explicit calculation in the model,
;; its code region is only 256 bytes long.
(defun code-start () #x10000000)
(defun code-size () (expt 2 8))
(defun code-end () (+ (code-start) (code-size)))
(defun guard-size () (expt 2 16))

;; Chunks in the model are 16 bytes long and 16-byte aligned, so the
;; low 4 bits of the first address in each chunk are zero.
(defun chunk-size-bits () 4)
(defun chunk-size () (expt 2 (chunk-size-bits)))

(defun code-region-p (addr)
  (and (addr-p addr) (>= addr (code-start)) (< addr (code-end))))


;; The code and data regions each have 'guard regions' directly before
;; and after them.  The presence of the guard regions ensures that
;; small offsets from legal addresses are also safe.
(defun code-guard-p (addr)
  (or (and (>= addr (code-end)) (< addr (+ (code-end) (guard-size))))
      (and (>= addr (- (code-start) (guard-size))) (< addr (code-start)))))

;; (loghead k n) is the k low bits of n, i.e. n & (2**k -1).
;; Thus, this checks whether the low 4 bits are all zero.
(defun chunk-aligned-p (addr)
  (= (loghead (chunk-size-bits) addr) 0))

;; In order to limit the use of the actual values of these constants,
;; we make them functions whose rules are disabled. For instance, this
;; makes it safe to write rewrite rules that target these values. If
;; you need a fact that can only be proved from the specific values,
;; you'll need to add a :in-theory enable hint.
(in-theory (disable code-start (code-start)))
(in-theory (disable code-size (code-size)))
(in-theory (disable code-end (code-end)))
(in-theory (disable chunk-size-bits (chunk-size-bits)))
(in-theory (disable chunk-size (chunk-size)))

;; The real system does not require that the code and data regions be
;; the same size, but it is convenient to have the data region be
;; small as well.
(defun data-start () #x20000000)
(defun data-size () (expt 2 8))
(defun data-end () (+ (data-start) (data-size)))

(defun data-region-p (addr)
  (and (addr-p addr) (>= addr (data-start)) (< addr (data-end))))

(defun data-guard-p (addr)
  (or (and (>= addr (data-end)) (< addr (+ (data-end) (guard-size))))
      (and (>= addr (- (data-start) (guard-size))) (< addr (data-start)))))

(in-theory (disable data-start (data-start)))
(in-theory (disable data-size (data-size)))
(in-theory (disable data-end (data-end)))

;; The 'zero-tag region' plays a role similar to the guard regions: it
;; allows the technique to enforce memory safety just by turning
;; address bits off.  It always starts at address 0, and must be at
;; least as large as either of the code or data regions.
(defun zero-tag-size () (expt 2 8))
(defun zero-tag-end () (zero-tag-size))

(defun zero-tag-region-p (addr)
  (and (addr-p addr) (>= addr 0) (< addr (zero-tag-end))))

;; One subtle point is that the zero tag region also requires its own
;; guard regions.  Because address arithmetic wraps around, the guard
;; region `before' the zero-tag region actually lies at the top of
;; memory.
(defun zero-tag-guard-p (addr)
  (or (and (>= addr (zero-tag-end)) (< addr (+ (zero-tag-end) (guard-size))))
      (and (>= addr (- (expt 2 32) (guard-size))) (< addr (expt 2 32)))))

;; Addition mod 32
(defun +-32 (x y) (loghead 32 (+ x y)))

;; Compute the length in bytes of the instruction starting at EIP; the
;; real x86 instruction encoding is used. See the definition of
;; regular-step for comments giving the mnemonics.
;;
;; Unknown instructions are treated as having length 1. An alternative
;; would be to return a distinguished value (such as nil) in this
;; case, but this would complicate many later definitions. It seems
;; sufficient to treat what is in fact a single unrecognized
;; instruction as a sequence of unrecognized ones: the safety check
;; willeventually reject such code in any case.
(defun insn-len (mem eip)
  (let ((byte1 (mem-fetch mem eip))
	(byte2 (mem-fetch mem (+-32 1 eip)))
	(byte3 (mem-fetch mem (+-32 2 eip))))
    (cond ((= byte1 #x90) 1)
	  ((and (= byte1 #x89) (= byte2 #xf6)) 2)
	  ((and (= byte1 #x8d) (= byte2 #x76) (= byte3 #x00)) 3)
	  ((= byte1 #xa1) 5)
	  ((= byte1 #xa3) 5)
	  ((= byte1 #xe9) 5)
	  ((= byte1 #xeb) 2)
	  ((and (= byte1 #x81) (= byte2 #xe3)) 6)
	  ((and (= byte1 #x81) (= byte2 #xe5)) 6)
	  ((= byte1 #x93) 1)
	  ((= byte1 #x95) 1)
	  ((and (= byte1 #x89) (= byte2 #x03)) 2)
	  ((and (= byte1 #x89) (= byte2 #x45) (= byte3 #x00)) 3)
	  ((and (= byte1 #xff) (= byte2 #xe3)) 2)
	  (t 1))))

;; Read a 32-bit little-endian word from the specified address, as an
;; unsigned integer
(defun fetch-word-le (mem addr)
  (+ (mem-fetch mem addr)
     (* (expt 2 8) (mem-fetch mem (+-32 addr 1)))
     (* (expt 2 16) (mem-fetch mem (+-32 addr 2)))
     (* (expt 2 24) (mem-fetch mem (+-32 addr 3)))))

;; Store a 32-bit word at the specified address
(defun store-word-le (mem addr word)
  (mem-store
   (mem-store
    (mem-store
     (mem-store mem addr (loghead 8 word))
     (+-32 addr 1) (loghead 8 (logtail 8 word)))
    (+-32 addr 2) (loghead 8 (logtail 16 word)))
   (+-32 addr 3) (loghead 8 (logtail 24 word))))

;; Interpret an unsigned byte as a signed byte
(defun sign-byte (byte)
  (if (<= byte #x7f)
      (nfix byte)
    (ifix (- byte #x100))))

;; Interpret an unsigned word as a signed word
(defun sign-word (word)
  (if (<= word #x7fffffff)
      (nfix word)
    (ifix (- word #x100000000))))

;; Bundle up the common checks that happen every time the model writes
;; to memory. Writing to a guard region is assumed to cause execution to
;; halt in a safe way, modelled by having the step function loop right
;; before the write. Writing any where else outside the data or zero-tag
;; region causes the execution to halt unsafely, modelled by the step
;; function returnning nil.
(defun do-write (x86 loop eaddr mem value)
  (cond ((or (data-region-p eaddr) (zero-tag-region-p eaddr))
	  (update-x86-state x86 :mem (store-word-le mem eaddr value)))
	((data-guard-p eaddr) loop)
	(t nil)))

;; Execute one step of the idealized processor, corresponding to a
;; single instruction. If the instruction is unsafe, returns nil. If
;; the instruction would in reality cause safe termination, it returns
;; the state unchanged. Otherwise, it returns the state after
;; execution of the instruction.
(defun regular-step (x86)
  (let* ((eip (x86-state-eip x86))
	 (mem (x86-state-mem x86))
	 (len (insn-len mem eip))
	 (next-eip (+-32 eip len))
	 (x86-2 (update-x86-state x86 :eip next-eip))
	 (insn1 (x86-state-fetch x86 eip))
	 (insn2 (x86-state-fetch x86 (+-32 1 eip)))
	 (insn3 (x86-state-fetch x86 (+-32 2 eip))))
    (cond
     ;; nop
     ((= insn1 #x90) x86-2)
     ;; 2-byte nop: mov %esi, %esi
     ((and (= insn1 #x89) (= insn2 #xf6)) x86-2)
     ;; 3-byte nop: lea 0x0(%esi),%esi
     ((and (= insn1 #x8d) (= insn2 #x76) (= insn3 #x00) x86-2))
     ;; inc %eax
     ((= insn1 #x40)
      (update-x86-state x86-2 :eax (+-32 1 (x86-state-eax x86))))
     ;; Direct read: mov addr, %eax
     ((= insn1 #xa1)
      (update-x86-state x86-2 :eax (fetch-word-le
				    mem 
				    (fetch-word-le mem (+-32 1 eip)))))
     ;; Direct write: mov %eax, addr
     ((= insn1 #xa3)
      (do-write x86-2 x86 (fetch-word-le mem (+-32 1 eip)) mem 
		(x86-state-eax x86)))
     ;; Direct jump: jmp addr (32-bit offset)
     ((= insn1 #xe9)
      (update-x86-state
       x86 :eip (+-32 next-eip
		      (sign-word (fetch-word-le mem (+-32 1 eip))))))
     ;; Direct jump: jmp addr (8-bit offset)
     ((= insn1 #xeb)
      (update-x86-state
       x86 :eip (+-32 next-eip
		      (sign-byte (mem-fetch mem (+-32 1 eip))))))
     ;; and <32-bit immed>, %ebx
     ((and (= insn1 #x81) (= insn2 #xe3))
      (update-x86-state x86-2 :ebx (logand (fetch-word-le mem (+-32 2 eip))
					   (x86-state-ebx x86-2))))
     ;; and <32-bit immed>, %ebp
     ((and (= insn1 #x81) (= insn2 #xe5))
      (update-x86-state x86-2 :ebp (logand (fetch-word-le mem (+-32 2 eip))
					   (x86-state-ebp x86-2))))
     ;; xchg %eax, %ebx
     ((= insn1 #x93)
      (update-x86-state x86-2 :eax (x86-state-ebx x86-2)
			:ebx (x86-state-eax x86-2)))
     ;; xchg %eax, %ebp
     ((= insn1 #x95)
      (update-x86-state x86-2 :eax (x86-state-ebp x86-2)
			:ebp (x86-state-eax x86-2)))
     ;; mov %eax, (%ebx)
     ((and (= insn1 #x89) (= insn2 #x03))
      (do-write x86-2 x86 (x86-state-ebx x86) mem 
		(x86-state-eax x86)))
     ;; mov %eax, (%ebp)
     ((and (= insn1 #x89) (= insn2 #x45) (= insn3 #x00))
      (do-write x86-2 x86 (x86-state-ebp x86) mem 
		(x86-state-eax x86)))
     ;; jmp *%ebx
     ((and (= insn1 #xff) (= insn2 #xe3))
      (update-x86-state x86 :eip (x86-state-ebx x86)))
     ;; Otherwise, crash
     (t nil))))

;; The complete processor step function includes a number of checks on
;; the program counter, which can cause execution to halt safely or
;; unsafely even before the instruction is decoded. (Put another way,
;; an unsafe jump causes execution to terminate not on the jump
;; instruction, but immediately before the target instruction; this
;; choice reduces the number of checks needed.)
(defun step-x86 (x86)
  (let* ((eip (x86-state-eip x86)))
    (cond ((code-guard-p eip) x86)
	  ((zero-tag-region-p eip) x86)
	  ((not (code-region-p eip)) nil)
	  (t (regular-step x86)))))

;; Return the k-th instruction among those that the sandboxing check
;; considers (when (code-start) is passed for "eip"). If k is large enough
;; that there is no kth instruction, returns nil.
(defun kth-insn-from (mem eip k)
  (declare (xargs :measure (acl2-count k)))
  (cond ((not (code-region-p eip)) nil)
	((zp k) eip)
	(t
	 (let ((prev (kth-insn-from mem eip (- k 1))))
	   (if (not (code-region-p prev))
	       nil
	     (let ((ret (+-32 prev (insn-len mem prev))))
	       (if (code-region-p ret) ret nil)))))))

;; Here and elsewhere, the suffix '-rec' refers to a recursive
;; function with an additional argument that implements a similarly
;; named function without the suffix.
(defun seq-reachable-rec (mem eip k)
  (if (zp k) (if (= eip (code-start)) 0 nil)
    (let ((kth-insn (kth-insn-from mem (code-start) k)))
    (or (and kth-insn (= eip kth-insn) k)
	(seq-reachable-rec mem eip (- k 1))))))

;; Is the instruction at eip among those that the sandboxing check
;; considers, namely one of the instructions in the sandboxed code
;; region that could be reached by a sequential decompilation starting
;; from (code-start)?
(defun seq-reachable-p (mem eip)
  (and (code-region-p eip)
       (not (null (seq-reachable-rec mem eip (code-size))))))

;; The static invariant that holds before the beginning of the code
(defun base-static-inv ()
  (static-inv nil nil t))

(defun check-ebx-data-inv (inv x86)
  (implies (static-inv-ebx-data-safe inv)
	   (or (data-region-p (x86-state-ebx x86))
	       (zero-tag-region-p (x86-state-ebx x86)))))

(defun check-ebx-code-inv (inv x86)
  (implies (static-inv-ebx-code-safe inv)
	   (and (chunk-aligned-p (x86-state-ebx x86))
		(or (code-region-p (x86-state-ebx x86))
		    (zero-tag-region-p (x86-state-ebx x86))))))

(defun check-ebp-inv (inv x86)
  (implies (static-inv-ebp-data-safe inv)
	   (or (data-region-p (x86-state-ebp x86))
	       (zero-tag-region-p (x86-state-ebp x86)))))
 
;; Check that a dynamic machine state satisfies the given static invariant.
(defun check-static-inv (inv x86)
  (and (check-ebx-data-inv inv x86)
       (check-ebx-code-inv inv x86)
       (check-ebp-inv inv x86)))

;; Turn off all the aspects of the static invariant that hold only for
;; the duration of a single instruction
(defun reset-static-inv (inv)
  (static-inv nil nil (static-inv-ebp-data-safe inv)))

;; Because the verification is sequential, the processor must be in a
;; relatively safe state before any jump. For instance, the frame
;; pointer %ebp must point at the data region.
(defun ok-to-jump (inv)
  (static-inv-ebp-data-safe inv))

;; Update the static invariant according to the effect of a single
;; instruction (this is the transition function if you think of it as
;; a state machine). Doesn't include the reset operation above.  This
;; function recognizes the three special AND instructions that are
;; used to sandbox pointers. It also recognizes the one instruction in
;; the implemented subset that can corrupt %ebp. Modifications of %ebx
;; do not need to be recognized because the safety invariants for %ebx
;; expire after a single instruction in any case.
(defun static-inv-trans (mem eip old-inv)
  (let ((byte1 (mem-fetch mem eip))
	(byte2 (mem-fetch mem (+-32 1 eip))))
    (cond ((and (= byte1 #x81) (= byte2 #xe3)
		(= (fetch-word-le mem (+-32 eip 2)) #x200000ff))
	   (update-static-inv old-inv :ebx-data-safe t))
	  ((and (= byte1 #x81) (= byte2 #xe3)
		(= (fetch-word-le mem (+-32 eip 2)) #x100000f0))
	   (update-static-inv old-inv :ebx-code-safe t))
	  ((and (= byte1 #x81) (= byte2 #xe5))
	   (update-static-inv old-inv :ebp-data-safe
			      (= (fetch-word-le mem (+-32 eip 2)) #x200000ff)))
	  ((= byte1 #x95)
	   (update-static-inv old-inv :ebp-data-safe nil))
	  (t old-inv))))

;; Compute the static invariant that must hold after the execution of
;; the kth instruction: conceptually, the invariants exists in the
;; points between instructions. Instructions k here and elsewhere
;; count from zero, so the invariant at the beginning of the code
;; corresponds to k = -1.
(defun static-inv-after-kth-insn (mem k)
  (declare (xargs :measure (acl2-count (+ k 1))))
  (cond ((not (integerp k)) nil)
	((< k -1) nil)
	((= k -1) (base-static-inv))
	((not (kth-insn-from mem (code-start) k)) nil)
	(t (static-inv-trans mem (kth-insn-from mem (code-start) k)
	    (reset-static-inv (static-inv-after-kth-insn mem (- k 1)))))))

;; Is the instruction at eip legal according to the local sandboxing
;; rules? (Local in the sense of not involving the static invariant;
;; peephole)
(defun insn-sandbox-ok-local (mem eip)
  (let ((byte1 (mem-fetch mem eip))
	(byte2 (mem-fetch mem (+-32 1 eip)))
	(byte3 (mem-fetch mem (+-32 2 eip)))
	;;(len (insn-len mem eip))
	)
    (cond ((= byte1 #x90) t)
	  ((and (= byte1 #x89) (= byte2 #xf6)) t)
	  ((and (= byte1 #x8d) (= byte2 #x76) (= byte3 #x00) t))
	  ((= byte1 #x40) t)
	  ((= byte1 #xa1) t)
	  ((= byte1 #xa3)
	   (data-region-p (fetch-word-le mem (+-32 eip 1))))
	  ((and (= byte1 #x81) (= byte2 #xe3) t))
	  ((and (= byte1 #x81) (= byte2 #xe5) t))
	  ((= byte1 #x93) t)
	  ((= byte1 #x95) t)
	  (t nil))))

;; As above, but add the rules that depend on the static invariant
(defun insn-sandbox-ok-inv (mem eip inv)
  (let ((byte1 (mem-fetch mem eip))
	(byte2 (mem-fetch mem (+-32 1 eip)))
	(byte3 (mem-fetch mem (+-32 2 eip)))
	(len (insn-len mem eip)))
    (cond ((and (= byte1 #x89) (= byte2 #x03))
	   (static-inv-ebx-data-safe inv))
	  ((and (= byte1 #x89) (= byte2 #x45) (= byte3 #x00))
	   (static-inv-ebp-data-safe inv))
	  ((and (= byte1 #xff) (= byte2 #xe3))
	   (and (ok-to-jump inv)
		(static-inv-ebx-code-safe inv)))
	  ((= byte1 #xe9)
	   (let ((eaddr (+-32 (+ eip len)
			      (sign-word (fetch-word-le mem (+-32 eip 1))))))
	     (and (ok-to-jump inv) 
		  (code-region-p eaddr) (chunk-aligned-p eaddr))))
	  ((= byte1 #xeb)
	   (let ((eaddr (+-32 (+ eip len)
			      (sign-byte (mem-fetch mem (+-32 eip 1))))))
	     (and (ok-to-jump inv)
		  (code-region-p eaddr) (chunk-aligned-p eaddr)))))))

;; This function converts from the representation of an instruction as
;; a pointer to its first byte (a program counter value) into its
;; index in the sequential order of instructions in the code
;; region. If the instruction is not in the sequence, returns nil.
(defun eip-to-k (mem eip)
  (seq-reachable-rec mem eip (code-size)))

;; The static invariant that must hold before an instruction is the
;; invariant that held after the last instruction, with transient
;; things expired if the instruction is aligned (might be a jump
;; target)
;;
;; Note that the use of reset-static-inv here is not redundant with
;; the use in static-inv-after-kth-insn: that one ensures that the
;; effect of a sandboxing instruction applies to at most one
;; subsequent instruction in any case, while this one ensures that it
;; applies to no subsequent instructions if the sandboxing instruction
;; appears at the end of a chunk.
(defun static-inv-before (mem eip)
  (let ((inv-after (static-inv-after-kth-insn
		    mem (- (eip-to-k mem eip) 1))))
    (if (chunk-aligned-p eip)
	(reset-static-inv inv-after)
      inv-after)))

;; The local and invariant-based parts of insn-sandbox-ok apply to
;; disjoint sets of instructions, but the default for unrecognized
;; instructions is false, so it is more convenient to combine them
;; with OR than with AND.
(defun insn-sandbox-ok (mem eip)
  (or (insn-sandbox-ok-local mem eip)
      (insn-sandbox-ok-inv mem eip (static-inv-before mem eip))))

(defun if-aligned-then-reach (mem addr)
  (or (not (code-region-p addr))
      (not (chunk-aligned-p addr))
      (seq-reachable-p mem addr)))

;; Check that each chunk-aligned address (up to the kth byte) in the
;; code region is sequentially reachable. Though somewhat inefficient,
;; counting by ones here makes the proof easier.
(defun all-aligned-reachable-rec (mem k)
  (if (zp k) (if-aligned-then-reach mem (code-start))
    (and (all-aligned-reachable-rec mem (- k 1))
	 (if-aligned-then-reach mem (+ (code-start) k)))))

(defun mem-sandbox-ok-rec (mem k)
  (if (zp k) (insn-sandbox-ok mem (code-start))
    (let ((addr (kth-insn-from mem (code-start) k)))
      (and (if (and addr (code-region-p addr))
	       (insn-sandbox-ok mem addr) t)
	   (mem-sandbox-ok-rec mem (- k 1))))))

;; Does the entire code region in the memory pass the static
;; sandboxing checks?
(defun mem-sandbox-ok (mem)
  (and (mem-sandbox-ok-rec mem (code-size))
       (all-aligned-reachable-rec mem (code-size))))

;; Execute for up to K instructions, returning nil if anything unsafe
;; happens along the way.
(defun step-for-k (x86 k)
  (declare (xargs :measure (acl2-count k)))
  (cond ((zp k) x86)
	(t (let ((k-minus-1 (step-for-k x86 (- k 1))))
	     (and k-minus-1
		  (step-x86 k-minus-1))))))

;; True if X86 is the sort of state that we guarantee will execute safely.
(defun is-start-state (x86)
  (and (x86-state-p x86)
       (equal (x86-state-eip x86) (code-start))
       (data-region-p (x86-state-ebp x86))
       (mem-sandbox-ok (x86-state-mem x86))))

;; These lemmas concern bitwise operations (conventionally named with
;; the prefix 'log' for `logical'), and give connections between
;; bitwise operators applied to words and arithmetic operators applied
;; to the values of those words as integers.

;; This seems to get ACL2 off into long pointless investigations of
;; rational numbers.
(in-theory (disable (:generalize mod-x-y-=-x+y)))

;; The first three lemmas are special cases of the distributivity of
;; bitwise and over bitwise or.
(defthm expand-data-mask
 (implies (natp addr)
	  (equal (logand addr (logior #x20000000 #x000000ff))
		 (logior (logand addr #x20000000)
			 (logand addr #x000000ff))))
 :rule-classes ((:rewrite :corollary
			  (implies (natp addr)
				   (equal (logand addr #x200000ff)
					  (logior (logand addr #x20000000)
						  (logand addr #x000000ff))))))
 :hints (("Goal" :in-theory (disable logior logand-with-mask
				     (binary-logior)))))

(defthm expand-code-mask-all-bits
 (implies (natp addr)
	  (equal (logand addr (logior #x10000000 #x000000ff))
		 (logior (logand addr #x10000000)
			 (logand addr #x000000ff))))
 :rule-classes ((:rewrite :corollary
			  (implies (natp addr)
				   (equal (logand addr #x100000ff)
					  (logior (logand addr #x10000000)
						  (logand addr #x000000ff))))))
 :hints (("Goal" :in-theory (disable logior logand-with-mask
				     (binary-logior)))))

(defthmd expand-code-mask-rounded
 (implies (natp addr)
	  (equal (logand addr (logior #x10000000 #x000000f0))
		 (logior (logand addr #x10000000)
			 (logand addr #x000000f0))))
 :rule-classes ((:rewrite :corollary
			  (implies (natp addr)
				   (equal (logand addr #x100000f0)
					  (logior (logand addr #x10000000)
						  (logand addr #x000000f0))))))
 :hints (("Goal" :in-theory (disable logior logand-with-mask
				     (binary-logior)))))

;; Facts about recursive functions in ACL2 are most easily proved if
;; the induction needed for the proof matches the structure of the
;; function's recursive definition. To take advantage of this, we
;; define several alternate versions of simple functions, and prove
;; that they are equivalent to ACL2's built-in definitions. A
;; particularly convenient perspective for bitwise operations is to
;; consider integers as lists of bits, least-significant first; ACL2
;; provides functions 'logcons', 'logcar', and 'logcdr' for this
;; purpose. The following function is equivalent to (expt 2 n).
(defun two-to-the-logcons (n)
  (if (zp n) 1
    (logcons 0 (two-to-the-logcons (- n 1)))))

(defun my-logand* (n m)
  (if (or (zp n) (zp m)) 0
    (logcons (b-and (logcar n) (logcar m))
	     (my-logand* (logcdr n) (logcdr m)))))

(defthm logcons-0-to-pos-is-pos
 (implies (posp n)
	  (< 0 (logcons 0 n))))

(defthm logcons-1-to-nat-is-pos
 (implies (natp n)
	  (< 0 (logcons 1 n))))

(defthm two-to-the-logcons-is-pos
  (and (< 0 (two-to-the-logcons k))
       (integerp (two-to-the-logcons k)))
  :rule-classes (:rewrite :type-prescription))

(defthm my-logand*-is-nonneg
  (and (<= 0 (my-logand* m n))
       (integerp (my-logand* m n)))
  :rule-classes (:rewrite :type-prescription))

(defun induct-logcdr-natural (bits k)
  (cond ((zp k) nil)
	((zp bits) nil)
	(t (induct-logcdr-natural (logcdr bits) (- k 1)))))

(defthm my-logand*-with-tttl-is-tttl-or-0-depending-on-logbitp
 (implies (and (natp addr)
	       (natp k))
	  (equal (my-logand* addr (two-to-the-logcons k))
		 (if (logbitp k addr) (two-to-the-logcons k) 0)))
 :hints (("Goal" :in-theory (set-difference-theories
			     (union-theories
			      (current-theory :here)
			      '(logops-recursive-definitions-theory logand*))
			     '(logand logcar logcdr logcons logbitp))
	  :induct (induct-logcdr-natural addr k))))

(defthmd rewrite-logand-as-my-logand*
 (implies (and (natp m) (natp n))
	  (equal (logand m n) (my-logand* m n)))
 :hints (("Goal" :induct (my-logand* m n))))

(defthmd rewrite-expt-2-as-two-to-the-logcons
  (implies (natp k)
	   (equal (expt 2 k) (two-to-the-logcons k))))

(defthm logand-with-bit-is-bit-or-0-depending-on-logbitp
 (implies (and (natp addr)
	       (natp k))
	  (equal (logand addr (expt 2 k))
		 (if (logbitp k addr) (expt 2 k) 0)))
 :hints (("Goal" :in-theory (enable rewrite-logand-as-my-logand*
				    rewrite-expt-2-as-two-to-the-logcons))))

(defthm logand-with-data-bit-is-bit-or-0-depending-on-logbitp
 (implies (natp addr)
	  (equal (logand addr #x20000000)
		 (if (logbitp 29 addr) #x20000000 0)))
 :hints (("Goal" :use
	  (:instance logand-with-bit-is-bit-or-0-depending-on-logbitp
		      (k 29)))))

(defthm logand-with-code-bit-is-bit-or-0-depending-on-logbitp
 (implies (natp addr)
	  (equal (logand addr #x10000000)
		 (if (logbitp 28 addr) #x10000000 0)))
 :hints (("Goal" :use
	  (:instance logand-with-bit-is-bit-or-0-depending-on-logbitp
		      (k 28)))))

(defun natural-induction (k)
  (if (zp k) nil (natural-induction (- k 1))))

(defun my-logor* (n m)
  (cond ((zp n) (nfix m))
	((zp m) (nfix n))
	(t (logcons (b-ior (logcar n) (logcar m))
		    (my-logor* (logcdr n) (logcdr m))))))

(defthm my-logor*-is-nonneg
  (and (<= 0 (my-logor* m n))
       (integerp (my-logor* m n)))
  :rule-classes (:rewrite :type-prescription))

(defun my-loghead* (k bits)
  (if (zp k) 0
    (logcons (logcar bits) (my-loghead* (- k 1) (logcdr bits)))))

(defthm my-loghead*-is-noneg
  (and (<= 0 (my-loghead* m n))
       (integerp (my-loghead* m n)))
  :rule-classes (:rewrite :type-prescription))

(defthmd rewrite-logior-as-my-logor*
 (implies (and (natp m) (natp n))
	  (equal (logior m n) (my-logor* m n)))
 :hints (("Goal" :in-theory (set-difference-theories
			     (union-theories
			      (current-theory :here)
			      '(logops-recursive-definitions-theory))
			     '(logior logcar logcdr logcons logbitp)))))

(defthmd rewrite-loghead-as-my-loghead*
  (implies (and (natp k) (natp bits))
           (equal (loghead k bits) (my-loghead* k bits)))
 :hints (("Goal" :in-theory (set-difference-theories
                             (union-theories
                              (current-theory :here)
                              '(logops-recursive-definitions-theory))
                             '(loghead logcar logcdr logcons logbitp)))))

(defthm logior-and-plus-same-way-beyond-loghead
 (implies (and (natp addr)
	       (natp bit)
	       (natp k))
	  (equal (logior (expt 2 (+ k bit)) (loghead bit addr))
		 (+ (expt 2 (+ k bit)) (loghead bit addr))))
 :hints (("Goal" :in-theory (set-difference-theories
                             (union-theories
                              (current-theory :here)
			      '(rewrite-logior-as-my-logor*
				rewrite-expt-2-as-two-to-the-logcons
				rewrite-loghead-as-my-loghead*))
			     '(logior)))))

(in-theory (disable logior))

(defthm logior-plus-loghead-special-case-1
 (implies (natp addr)
	  (equal (logior (expt 2 29) (loghead 8 addr))
		 (+ (expt 2 29) (loghead 8 addr))))
 :hints (("Goal" :in-theory (disable (expt) expt)
	  :use (:instance logior-and-plus-same-way-beyond-loghead
			  (bit 8) (k 21))))
 :rule-classes
 ((:rewrite :corollary
	    (implies (natp addr)
		     (equal (logior #x20000000 (loghead 8 addr))
			    (+ #x20000000(loghead 8 addr)))))))

(defthm logior-plus-loghead-special-case-2
 (implies (natp addr)
	  (equal (logior (expt 2 28) (loghead 8 addr))
		 (+ (expt 2 28) (loghead 8 addr))))
 :hints (("Goal" :in-theory (disable (expt) expt)
	  :use (:instance logior-and-plus-same-way-beyond-loghead
			  (bit 8) (k 20))))
 :rule-classes
 ((:rewrite :corollary
	    (implies (natp addr)
		     (equal (logior #x10000000 (loghead 8 addr))
			    (+ #x10000000(loghead 8 addr)))))))

(defthm my-logand*-w-aligned-mask
 (implies (and (natp addr)
	       (equal (my-loghead* bits mask) 0))
	  (equal (my-loghead* bits (my-logand* mask addr)) 0)))

(defthm logand-w-aligned-mask
 (implies (and (natp addr)
	       (natp mask)
	       (natp bits)
	       (equal (loghead bits mask) 0))
	  (equal (loghead bits (logand mask addr)) 0))
 :hints (("Goal" :in-theory (e/d
			     (rewrite-logand-as-my-logand*
			      rewrite-loghead-as-my-loghead*)
			     (logand loghead)))))

(defthm code-mask-is-aligned
 (equal (loghead (chunk-size-bits) #x100000f0) 0)
 :hints (("Goal" :in-theory (enable chunk-size-bits))))

(defun loground (bits n)
  (if (zp bits) n
    (logcons 0 (loground (- bits 1) (logcdr n)))))

;; While unsigned integers can be considered as lists of bits with
;; infinitely many zeros at the most-significant tail, signed integers
;; can be considered in two's complement as having a tail of either 0s
;; are 1s (0 or -1).
(defun my-lognot* (n)
  (cond ((zip n) -1)
	((= n -1) 0)
	(t (logcons (b-not (logcar n)) (my-lognot* (logcdr n))))))

(defun my-logand2 (n m)
  (cond ((zip n) 0)
	((zip m) 0)
	((= n -1) m)
	((= m -1) n)
	(t (logcons (b-and (logcar n) (logcar m))
		    (my-logand2 (logcdr n) (logcdr m))))))

(defun my-logmask* (size)
  (if (zp size) 0
    (logcons 1 (my-logmask* (- size 1)))))

;; (thm
;;  (equal (my-logmask* size)
;; 	(- (two-to-the-logcons size) 1)))

(defthm my-logand2-of-my-lognot*-of-my-logmask*-is-loground
 (implies (and (natp n) (natp bits))
	  (equal (my-logand2 (my-lognot* (my-logmask* bits)) n)
		 (loground bits n)))
 :hints (("Goal" :in-theory (disable logcar logcons logcdr))))

(defthm my-logand2-of-my-logmask*-is-loghead
 (implies (and (natp addr)
	       (natp k))
	  (equal (my-logand2 (my-logmask* k) addr)
		 (my-loghead* k addr)))
 :hints (("Goal" :in-theory (disable logcar logcons logcdr))))

(defthm my-logand2-w-partial-mask-is-extract-bitfield
 (implies (and (natp addr)
	       (natp k) (natp j))
	  (equal (my-logand2 (my-lognot* (my-logmask* j))
			     (my-logand2 (my-logmask* k)
					 addr))
		 (loground j (my-loghead* k addr))))
 :hints (("Goal" :in-theory (disable logcar logcons logcdr))))

(defthm associativity-of-my-logand2
 (implies (and (integerp a) (integerp b) (integerp c))
	  (equal (my-logand2 a (my-logand2 b c))
		 (my-logand2 (my-logand2 a b) c)))
 :hints (("Goal" :in-theory (disable logcar logcdr logcons)
	  :induct (logcdr-induction-3 a b c))))

(defthm my-logand2-with-240-is-4-bits
 (implies (natp addr)
	  (equal (my-logand2 #xf0 addr)
		 (loground 4 (my-loghead* 8 addr))))
 :hints (("Goal" :use (:instance my-logand2-w-partial-mask-is-extract-bitfield
				 (j 4) (k 8)))))

(defthmd rewrite-logand-as-my-logand2
 (implies (and (natp m) (natp n))
	  (equal (logand m n) (my-logand2 m n)))
 :hints (("Goal" :induct (my-logand2 m n))))


(defthm logand-with-240-is-4-bits
 (implies (natp addr)
	  (equal (logand #xf0 addr)
		 (loground 4 (loghead 8 addr))))
 :hints (("Goal" :in-theory (enable rewrite-logand-as-my-logand2
				    rewrite-loghead-as-my-loghead*))))

(defthm commute-my-loghead*-with-loground
 (implies (and (natp addr) (natp k) (natp j))
	  (equal (loground j (my-loghead* k addr))
		 (my-loghead* k (loground j addr)))))

(defthm loground-of-nat-is-nat
 (implies (and (natp addr)
	       (natp j))
	  (natp (loground j addr)))
 :rule-classes (:rewrite :type-prescription))

(defthm commute-loghead-with-loground
 (implies (and (natp addr) (natp k) (natp j))
	  (equal (loground j (loghead k addr))
		 (loghead k (loground j addr))))
 :hints (("Goal" :in-theory (enable rewrite-loghead-as-my-loghead*)
	  :do-not-induct t)))

(defthm logand-w-code-mask-region-cases
 (implies (and (natp addr)
	       (not (equal (logand addr #x100000f0)
			   (loground 4 (loghead 8 addr)))))
	  (equal (logand addr #x100000f0)
		 (+ #x10000000 (loground 4 (loghead 8 addr)))))
 :hints (("Goal" :do-not-induct t
	  :in-theory (e/d (expand-code-mask-rounded)
			  (logior logbitp))
	  :cases ((equal (logior (expt 2 28) (loground 4 (loghead 8 addr)))
			 (+ (expt 2 28) (loground 4 (loghead 8 addr))))))))

;; The intuitive statement of the following lemma would be a
;; disjunction, but such a lemma would not work well as a rewrite
;; rule, so we arbitrarily choose one disjunct to be the conclusion
;; and rewrite it as an implication.
(defthm logand-w-code-mask-is-code-or-zt
  (implies (and (natp addr)
		(not (zero-tag-region-p (logand addr #x100000f0))))
	   (code-region-p (logand addr #x100000f0)))
  :hints (("Goal" :do-not-induct t
	   :use (:instance logand-w-code-mask-region-cases)
	   :in-theory (enable code-start code-end code-size))))
	   
;;; Type-like lemmas

(in-theory (disable mem-fetch code-region-p data-region-p))
(in-theory (disable loghead logtail +-32))
(in-theory (disable sign-byte sign-word))

(defthm fetch-byte-is-byte
 (implies (mem-p mem)	  
	  (and 
	   (integerp (mem-fetch mem addr))
	   (>= (mem-fetch mem addr) 0)
	   (< (mem-fetch mem addr) 256)))
 :hints (("Goal" :in-theory (enable mem-fetch bound?-equal
				    binding-equal)))
 :rule-classes (:rewrite :type-prescription))

(defthm x86-mem-is-mem
  (implies (x86-state-p x86)
	   (mem-p (x86-state-mem x86))))

(defthm fetch-byte-from-mem-is-byte
 (implies (x86-state-p x86)	  
	  (and 
	   (integerp (mem-fetch (x86-state-mem x86) addr))
	   (>= (mem-fetch (x86-state-mem x86) addr) 0)
	   (< (mem-fetch (x86-state-mem x86) addr) 256)))
 :rule-classes (:rewrite :type-prescription))

;; 256, 65536, 16777216, and 4294967296 are 2**8, 2**16, 2**24, and
;; 2**32 but they cannot be written as (expt 2 8) (etc.) in the
;; statement of a theorem, because this would prevent the generated
;; rewrite rule from ever being applicable.
(defthm assemble-bytes-is-word
 (let ((word (+ byte1 (* 256 byte2) (* 65536 byte3)
		(* 16777216 byte4))))
   (implies (and (byte-p byte1) (byte-p byte2)
		 (byte-p byte3) (byte-p byte4))
	    (and (integerp word)
		 (>= word 0)
		 (< word 4294967296)))))

(defthm fetch-word-is-word
 (implies (mem-p mem)
	  (and (integerp (fetch-word-le mem addr))
	       (>= (fetch-word-le mem addr) 0)
	       (< (fetch-word-le mem addr) 4294967296)))
 :rule-classes (:rewrite :type-prescription))

(defthm fetch-word-from-mem-is-word
 (implies (x86-state-p x86)
	  (and (integerp (fetch-word-le (x86-state-mem x86) eip))
	       (<= 0 (fetch-word-le (x86-state-mem x86) eip))
	       (< (fetch-word-le (x86-state-mem x86) eip) 4294967296)))
 :rule-classes (:rewrite :type-prescription))

(defthm +-32-of-ints-is-addr
 (implies (and (integerp a1) (integerp a2))
	  (and (integerp (+-32 a1 a2))
	       (<= 0 (+-32 a1 a2))
	       (< (+-32 a1 a2) 4294967296)
	       (addr-p (+-32 a1 a2))))
 :hints (("Goal" :in-theory (enable +-32))))

(defthm store-word-is-mem
 (implies (and (mem-p mem)
	       (addr-p addr)
	       (addr-p word))
	  (mem-p (store-word-le mem addr word))))

(in-theory (disable fetch-word-le store-word-le))

(defthm and-of-nonneg-is-nonneg
 (implies (and (natp n) (integerp m))
	  (<= 0 (logand n m)))
 :hints (("Goal" :use (:instance rewrite-logand-as-my-logand*))))

(defthm x86-state-int-regs
 (implies (x86-state-p x86-state)
	  (and (integerp (x86-state-eax x86-state))
	       (integerp (x86-state-ebx x86-state))
	       (integerp (x86-state-ecx x86-state))
	       (integerp (x86-state-edx x86-state))
	       (integerp (x86-state-esi x86-state))
	       (integerp (x86-state-edi x86-state))
	       (integerp (x86-state-ebp x86-state))
	       (integerp (x86-state-esp x86-state))
	       (integerp (x86-state-eip x86-state))
	       (integerp (x86-state-eflags x86-state)))))

(defthm x86-state-nonneg-regs
 (implies (x86-state-p x86-state)
	  (and (<= 0 (x86-state-eax x86-state))
	       (<= 0 (x86-state-ebx x86-state))
	       (<= 0 (x86-state-ecx x86-state))
	       (<= 0 (x86-state-edx x86-state))
	       (<= 0 (x86-state-esi x86-state))
	       (<= 0 (x86-state-edi x86-state))
	       (<= 0 (x86-state-ebp x86-state))
	       (<= 0 (x86-state-esp x86-state))
	       (<= 0 (x86-state-eip x86-state))
	       (<= 0 (x86-state-eflags x86-state)))))

(defthm x86-regs-bound
 (implies (x86-state-p x86-state)
	  (and (< (x86-state-eax x86-state) 4294967296)
	       (< (x86-state-ebx x86-state) 4294967296)
	       (< (x86-state-ecx x86-state) 4294967296)
	       (< (x86-state-edx x86-state) 4294967296)
	       (< (x86-state-esi x86-state) 4294967296)
	       (< (x86-state-edi x86-state) 4294967296)
	       (< (x86-state-ebp x86-state) 4294967296)
	       (< (x86-state-esp x86-state) 4294967296)
	       (< (x86-state-eip x86-state) 4294967296)
	       (< (x86-state-eflags x86-state) 4294967296))))

(in-theory (disable insn-sandbox-ok-local insn-sandbox-ok-inv
		    insn-sandbox-ok insn-len static-inv-trans))

(defthm reg-step-exists-implies-int-regs
 (implies (and (x86-state-p x86-state)
	       (regular-step x86-state))
	  (and (integerp (x86-state-eax (regular-step x86-state)))
	       (integerp (x86-state-ebx (regular-step x86-state)))
	       (integerp (x86-state-ecx (regular-step x86-state)))
	       (integerp (x86-state-edx (regular-step x86-state)))
	       (integerp (x86-state-esi (regular-step x86-state)))
	       (integerp (x86-state-edi (regular-step x86-state)))
	       (integerp (x86-state-ebp (regular-step x86-state)))
	       (integerp (x86-state-esp (regular-step x86-state)))
	       (integerp (x86-state-eip (regular-step x86-state)))
	       (integerp (x86-state-eflags (regular-step x86-state))))))

(defthm state-p-type-preservation
 (implies (and (x86-state-p x86-state)
               (step-x86 x86-state))
          (x86-state-p (step-x86 x86-state))))

;;; Properties of instruction counting

;; Instructions in the code region can be identified by two numbering
;; schemes: the processor itself uses the address of the first byte of
;; the instruction (variables named 'eip', after the program counter
;; register), but for tasks like computing the predecessor of an
;; instruction, it more useful to number the instructions sequentially
;; from 0 (as in the argument 'k' to kth-insn-from). Because of the
;; possibility of overlapping instructions, such a sequential
;; numbering may not exist in arbitrary x86 code, but it does exist
;; for our system as long as only 'sequentially reachable'
;; instructions are considered, those instructions that would be
;; encountered on a sequential disassembly starting at the beginning
;; of the code region. (In the names of functions and theorems,
;; `sequentially reachable' is abbreviated 'seq-reach').

;; Periodically during development, we used a script to check whether
;; the proof could still be verified with any lemmas removed. If so,
;; we usually removed them; besides the aesthetic benefits of
;; simplification, this sometimes significantly improved ACL2's
;; running time, by pruning unproductive parts of its proof
;; search. Earlier versions of the proof had '-1', '-2', and '-3'
;; variants of this lemma, but they turned out to be unneeded.
(defthm code-step-doesnt-wrap-4-lh
 (implies (and (natp k)
	       (addr-p addr)
	       (code-region-p addr)
	       (<= k #x10000))
	  (equal (+-32 addr k) (+ addr k)))
 :hints (("Goal" :do-not-induct t
	  :in-theory (enable code-region-p code-start code-end code-size
			     +-32))))

;; Same as above, but with the args to "+" commuted. This is needed
;; because if one of the things being added is abstract, ACL2 doesn't
;; always try permutations when checking rule applicability
(defthm code-step-doesnt-wrap-4c-lh
 (implies (and (natp k)
	       (addr-p addr)
	       (code-region-p addr)
	       (<= k #x10000))
	  (equal (+-32 k addr) (+ k addr)))
 :hints (("Goal" :do-not-induct t
	  :in-theory (enable code-region-p code-start code-end code-size
			     +-32))))

(defthm code-start-code-region
 (code-region-p (code-start))
 :hints (("Goal" :in-theory (enable code-start))))

(defthm kth-insn-is-code-from-start
 (implies (kth-insn-from mem (code-start) k)
	  (code-region-p (kth-insn-from mem (code-start) k))))

(defthm in-code-is-address
 (implies (code-region-p addr)
	  (addr-p addr))
 :hints (("Goal" :in-theory (enable code-region-p))))

;;  The purpose of the following lemma is to provide a bound on the
;;  length of an instruction that can be used in proofs where we do
;;  not wish to expand the definition of 'insn-len' (since it consists
;;  of many cases).
(defthm weak-insn-len-bound
 (<= (insn-len mem insn) 65536)
 :hints (("Goal" :in-theory (enable insn-len))))

(defthm seq-reach-of-next-insn-w-next-bound-nowrap
 (let ((next (+ eip (insn-len mem eip))))
   (implies (and (mem-p mem)
		 (addr-p eip)
		 (natp k)
		 (code-region-p eip)
		 (code-region-p next)
		 (seq-reachable-rec mem eip k))
	    (seq-reachable-rec mem next (+ 1 k))))
 :hints (("Goal" :induct (seq-reachable-rec mem eip k))))

;; "ito" = "in terms of"
(defthm kth-insn-loc-ito-k
 (implies (and (mem-p mem)
	       (natp k)
	       (kth-insn-from mem (code-start) k))
	  (and
	   (>= (kth-insn-from mem (code-start) k)
	       (+ (code-start) k))
	   (< (kth-insn-from mem (code-start) k)
	      (code-end))))
 :hints (("Goal" :in-theory (enable code-start code-size code-end
				    code-region-p))))

(defthm kth-insn-loc-ito-k-beyond-end
 (implies (and (mem-p mem)
	       (natp k)
	       (not (and
		     (>= (kth-insn-from mem (code-start) k)
			 (+ (code-start) k))
		     (< (kth-insn-from mem (code-start) k)
			(code-end)))))
	  (not (kth-insn-from mem (code-start) k))))

(defthm no-such-beyond-end
  (implies (and (mem-p mem)
		(natp k)
		(>= k (code-size))
		(>= (kth-insn-from mem (code-start) k)
		    (+ (code-start) k)))
	   (not (< (kth-insn-from mem (code-start) k)
		   (code-end))))
 :hints (("Goal" :in-theory (enable code-start code-size code-end
				    code-region-p))))

;; The development of this lemma (using the three preceding ones)
;; seemed awkward because of the negation, but we were not able to
;; find a more direct proof method.
(defthm no-insns-beyond-size
 (implies (and (mem-p mem)
	       (natp k)
	       (>= k (code-size)))
	  (not (kth-insn-from mem (code-start) k)))
 :hints (("Goal" :do-not-induct t)))

(defthm seq-reach-next-with-size-nowrap
 (let ((next (+ eip (insn-len mem eip))))
   (implies (and (mem-p mem)
		 (addr-p eip)
		 (code-region-p eip)
		 (code-region-p next)
		 (natp (- (code-size) 1))
		 (seq-reachable-rec mem eip (code-size)))
	    (seq-reachable-rec mem next (code-size))))
 :hints (("Goal" :in-theory (disable seq-reachable-rec)
	  :cases ((seq-reachable-rec
		   mem (+-32 eip (insn-len mem eip))
		   (+ 1 (code-size)))))
	 ("Subgoal 1" :in-theory
	  (disable seq-reach-of-next-insn-w-next-bound-nowrap))))

(defthm all-aligned-reach-and-aligned-is-reach-by-k
 (implies (and (natp k)
	       (all-aligned-reachable-rec mem k)
	       (> k (- addr (code-start)))
	       (code-region-p addr)
	       (chunk-aligned-p addr))
	  (seq-reachable-rec mem addr (code-size)))
 :hints (("Goal" :in-theory (enable code-start code-size chunk-size
				    chunk-size-bits code-region-p))))

(defthm all-aligned-reach-and-aligned-is-reach
 (implies (and (all-aligned-reachable-rec mem (code-size))
	       (code-region-p addr)
	       (chunk-aligned-p addr))
	  (seq-reachable-rec mem addr (code-size)))
 :hints (("Goal" :use (:instance all-aligned-reach-and-aligned-is-reach-by-k
				 (k (code-size)))
	  :in-theory (enable code-start code-size code-end code-region-p)
	  :do-not-induct t)))

;;; Dynamic invariants

;; With the preliminaries above aside, we now introduce the
;; definitions for the main safety invariant, and prove a few of its
;; simple properties.

(in-theory (disable step-x86 regular-step))

(defun safe-loop (x86-state)
  (or (code-guard-p (x86-state-eip x86-state))
      (zero-tag-region-p (x86-state-eip x86-state))))

;; The structure of the invariant is slightly more complicated than a
;; simple conjunction, because of the treatment of jumps to the code
;; guard or zero-tag regions. Recall that such jumps are safely
;; trapped, which is modelled by an infinite loop in the simulated
;; processor, but that the loop conceptually occurs after the jump,
;; when the program counter is pointing at the guard region. This
;; means the proof cannot guarantee that the program counter will, for
;; instance, always stay in the code region, as this property would
;; not hold during a safe loop. The invariant is therefore divided
;; into two definitions: 'regular-invariant' includes those properties
;; that hold whenever code is executing normally, and
;; 'safety-invariant' requires either that they hold, or that the code
;; is in a safe loop. In addition, 'safety-invariant' includes some
;; other properties that hold regardless of the program counter's
;; location.
(defun regular-invariant (x86-state)
  (let ((eip (x86-state-eip x86-state))
	(mem (x86-state-mem x86-state)))
    (and (code-region-p eip)
	 (seq-reachable-p mem eip)
	 (insn-sandbox-ok mem eip)
	 (check-static-inv (static-inv-before mem eip) x86-state))))

(defun safety-invariant (x86-state)
  (and (x86-state-p x86-state)
       (mem-sandbox-ok (x86-state-mem x86-state))
       (or (safe-loop x86-state)
	   (regular-invariant x86-state))))
 
(in-theory (disable x86-state-p))

;; Recall that insn-sandbox-ok-local is true only for those
;; instructions whose safety does not depend on the static invariant:
;; for instance, these two lemmas do not say anything about the safety
;; of an indirect write.
(defthm local-sb-ok-implies-safe-reg-step
 (implies (and (x86-state-p x86)
	       (seq-reachable-rec (x86-state-mem x86) (x86-state-eip x86)
				  (code-size))
	       (code-region-p (x86-state-eip x86))
	       (insn-sandbox-ok-local (x86-state-mem x86)
				      (x86-state-eip x86)))
	  (regular-step x86))
 :hints (("Goal" :in-theory (enable regular-step
				    insn-sandbox-ok
				    insn-sandbox-ok-local insn-sandbox-ok-inv
				    data-region-p data-start
				    data-size data-end code-region-p))))

(defthm local-sb-ok-implies-safe-step
 (implies (and (x86-state-p x86)
	       (seq-reachable-rec (x86-state-mem x86) (x86-state-eip x86)
				  (code-size))
	       (code-region-p (x86-state-eip x86))
	       (insn-sandbox-ok-local (x86-state-mem x86)
				      (x86-state-eip x86)))
	  (step-x86 x86))
 :hints (("Goal" :in-theory (enable step-x86
				    insn-sandbox-ok
				    insn-sandbox-ok-local insn-sandbox-ok-inv
				    data-region-p data-start
				    data-size data-end code-region-p))))

(defthmd regular-inv-implies-safe-reg-step-short-jump
 (implies (and (x86-state-p x86-state)
	       (mem-sandbox-ok-rec (x86-state-mem x86-state) (code-size))
	       (code-region-p (x86-state-eip x86-state))
	       (equal (mem-fetch (x86-state-mem x86-state)
				 (x86-state-eip x86-state))
		      235)
	       (regular-invariant x86-state))
	  (regular-step x86-state))
 :hints (("Goal" :in-theory (e/d (regular-step insn-sandbox-ok
					       insn-sandbox-ok-inv)
				 (check-static-inv)))))

(defthmd regular-inv-implies-safe-reg-step-long-jump
 (implies (and (x86-state-p x86-state)
	       (mem-sandbox-ok-rec (x86-state-mem x86-state) (code-size))
	       (code-region-p (x86-state-eip x86-state))
	       (equal (mem-fetch (x86-state-mem x86-state)
				 (x86-state-eip x86-state))
		      233)
	       (regular-invariant x86-state))
	  (regular-step x86-state))
 :hints (("Goal" :in-theory (e/d (regular-step insn-sandbox-ok
					       insn-sandbox-ok-inv)
				 (check-static-inv)))))

(defthmd regular-inv-implies-safe-reg-step-non-jump
 (implies (and (x86-state-p x86-state)
	       (mem-sandbox-ok-rec (x86-state-mem x86-state) (code-size))
	       (code-region-p (x86-state-eip x86-state))
	       (not (equal (mem-fetch (x86-state-mem x86-state)
				      (x86-state-eip x86-state))
			   233))
	       (not (equal (mem-fetch (x86-state-mem x86-state)
				      (x86-state-eip x86-state))
			   235))
	       (regular-invariant x86-state))
	  (regular-step x86-state))
 :hints (("Goal" :in-theory (enable regular-step insn-sandbox-ok
				    insn-sandbox-ok-inv))))

(defthm regular-inv-implies-safe-reg-step
 (implies (and (x86-state-p x86-state)
	       (mem-sandbox-ok-rec (x86-state-mem x86-state) (code-size))
	       (code-region-p (x86-state-eip x86-state))
	       (regular-invariant x86-state))
	  (regular-step x86-state))
 :hints (("Goal" :in-theory
	  (union-theories
	   (theory 'minimal-theory)
	   '(regular-inv-implies-safe-reg-step-short-jump
	     regular-inv-implies-safe-reg-step-long-jump
	     regular-inv-implies-safe-reg-step-non-jump))
	  :cases ((equal (mem-fetch (x86-state-mem x86-state)
				    (x86-state-eip x86-state))
			 233)
		  (equal (mem-fetch (x86-state-mem x86-state)
				    (x86-state-eip x86-state))
			 235)))))

(defthm regular-inv-implies-safe-step
 (implies (and (x86-state-p x86-state)
	       (mem-sandbox-ok-rec (x86-state-mem x86-state) (code-size))
	       (regular-invariant x86-state))
	  (step-x86 x86-state))
 :hints (("Goal" :in-theory
	  (union-theories (theory 'minimal-theory)
			  '(step-x86 regular-invariant x86-state-p
				     weak-x86-state-p
				     regular-inv-implies-safe-reg-step)))))
			  
(defthm step-from-code-is-guard-or-code-nowrap
 (let ((next (+ eip (insn-len mem eip))))
   (implies (and (code-region-p eip)
		 (mem-p mem)
		 (not (code-guard-p next)))
	    (code-region-p next)))
 :hints (("Goal" :in-theory (enable code-region-p code-start
				    code-size code-end insn-len))))

(in-theory (disable code-region-p (code-region-p)
		    guard-size (guard-size)))

(defthm seq-reach-is-in-code
 (implies (seq-reachable-rec mem eip k)
	  (code-region-p eip)))

;; This rule is annoying to have enabled because it tends to eliminate
;; insn-sandbox-ok from hypotheses, which prevents it from being
;; expanded.
(defthm seq-reach-in-sb-is-ok
 (implies (and (mem-sandbox-ok-rec mem k)
	       (seq-reachable-rec mem eip k))
	  (insn-sandbox-ok mem eip)))
(in-theory (disable seq-reach-in-sb-is-ok))

(in-theory (disable code-guard-p))

;;; Data modification irrelevance

;; These lemmas are devoted to the fact that writing to one of the
;; legal places to write (the data section or the zero-tag region)
;; does not affect the results of the checks performed on the code
;; region. Though intuitively a simple consequence of the fact that
;; the code and data regions are disjoint, this is somewhat cumbersome
;; to prove because both regions are represented in the same memory
;; data structure.

(defthm mem-fetch-code-same-after-data-store
 (implies (and (mem-p mem)
	       (code-region-p eip)
	       (integerp addr)
	       (or (data-region-p addr) (zero-tag-region-p addr))
	       (addr-p word))
	  (equal (mem-fetch (store-word-le mem addr word) eip)
		 (mem-fetch mem eip)))
 :hints (("Goal" :in-theory (enable code-start data-start code-region-p
				    data-region-p code-size data-size
				    insn-len mem-fetch store-word-le
				    code-end data-end +-32))))

(defthm mem-fetch-code-plus-k-same-after-data-store
 (implies (and (mem-p mem)
	       (natp k)
	       (< k 65536)
	       (code-region-p eip)
	       (integerp addr)
	       (or (data-region-p addr) (zero-tag-region-p addr))
	       (addr-p word))
	  (equal (mem-fetch (store-word-le mem addr word) (+ k eip))
		 (mem-fetch mem (+ k eip))))
 :hints (("Goal" :in-theory (enable code-start data-start code-region-p
				    data-region-p code-size data-size
				    insn-len mem-fetch store-word-le
				    code-end data-end +-32))))

(defthm mem-fetch-code-plus-k-c-same-after-data-store
 (implies (and (mem-p mem)
	       (natp k)
	       (< k 65536)
	       (code-region-p eip)
	       (integerp addr)
	       (or (data-region-p addr) (zero-tag-region-p addr))
	       (addr-p word))
	  (equal (mem-fetch (store-word-le mem addr word) (+ eip k))
		 (mem-fetch mem (+ eip k))))
 :hints (("Goal" :in-theory (enable code-start data-start code-region-p
				    data-region-p code-size data-size
				    insn-len mem-fetch store-word-le
				    code-end data-end +-32))))

(defthm mem-fetch-code-plus32-k-j-same-after-data-store
 (implies (and (mem-p mem)
	       (natp k)
	       (< k 65536)
	       (natp j)
	       (< j 65536)
	       (code-region-p eip)
	       (integerp addr)
	       (or (data-region-p addr) (zero-tag-region-p addr))
	       (addr-p word))
	  (equal (mem-fetch (store-word-le mem addr word)
			    (+-32 (+ eip j) k))
		 (mem-fetch mem (+-32 (+ eip j) k))))
 :hints (("Goal" :in-theory (enable code-start data-start code-region-p
				    data-region-p code-size data-size
				    insn-len mem-fetch store-word-le
				    code-end data-end +-32))))

(defthm nil-not-in-code-region
  (not (code-region-p nil))
  :hints (("Goal" :in-theory (enable code-region-p))))

(defthm insn-len-same-after-data-store
  (implies (and (mem-p mem)
		(code-region-p eip)
		(integerp addr)
		(or (data-region-p addr) (zero-tag-region-p addr))
		(addr-p word))
	   (equal (insn-len (store-word-le mem addr word) eip)
		  (insn-len mem eip)))
  :hints (("Goal" :in-theory (enable insn-len))))
		
(defthm kth-insn-same-after-data-store
 (implies (and (mem-p mem)
	       (natp k)
	       (or (data-region-p addr) (zero-tag-region-p addr))
	       (addr-p word))
	  (equal (kth-insn-from (store-word-le mem addr word) (code-start) k)
		 (kth-insn-from mem (code-start) k)))
 :hints (("Goal" :in-theory (enable code-start data-region-p))))

(defthm fetch-word-code-plus-k-same-after-data-store
 (implies (and (mem-p mem)
	       (natp k)
	       (< k 65530)
	       (code-region-p eip)
	       (integerp addr)
	       (or (data-region-p addr) (zero-tag-region-p addr))
	       (addr-p word))
	  (equal (fetch-word-le (store-word-le mem addr word) (+ k eip))
		 (fetch-word-le mem (+ k eip))))
 :hints (("Goal" :in-theory (enable code-start data-start code-region-p
				    data-region-p code-size data-size
				    data-end fetch-word-le)
	  :do-not-induct t)))

(defthm static-inv-trans-same-after-data-store
 (implies (and (mem-p mem)
	       (integerp addr)
	       (or (data-region-p addr) (zero-tag-region-p addr))
	       (code-region-p eip)
	       (static-inv-p inv)
	       (addr-p word))
	  (equal (static-inv-trans (store-word-le mem addr word) eip inv)
		 (static-inv-trans mem eip inv)))
 :hints (("Goal" :in-theory (enable static-inv-trans))))

(defthm static-inv-trans-is-inv
 (implies (static-inv-p inv)
	  (static-inv-p (static-inv-trans mem eip inv)))
 :hints (("Goal" :in-theory (enable static-inv-trans))))

(defthm s-i-ebp-data-safe-is-bool
 (implies (or (static-inv-p inv) (not inv))
	  (booleanp (static-inv-ebp-data-safe inv)))
 :rule-classes (:rewrite :type-prescription))

(defthm s-i-ebp-data-safe-of-after-kth-is-bool
 (booleanp (static-inv-ebp-data-safe
	    (static-inv-after-kth-insn mem k))))

(defthm static-inv-after-kth-same-after-data-store
 (implies (and (mem-p mem)
	       (integerp addr)
	       (or (data-region-p addr) (zero-tag-region-p addr))
	       (addr-p word))
	  (equal (static-inv-after-kth-insn (store-word-le mem addr word) k)
		 (static-inv-after-kth-insn mem k))))

(defthm seq-reach-same-after-data-store
 (implies (and (mem-p mem)
	       (integerp addr)
	       (or (data-region-p addr) (zero-tag-region-p addr))
	       (addr-p word))
	  (equal (seq-reachable-rec (store-word-le mem addr word) eip k)
		 (seq-reachable-rec mem eip k))))

(defthm static-inv-before-same-after-data-store
 (implies (and (mem-p mem)
	       (addr-p eip)
	       (addr-p addr)
	       (or (data-region-p addr) (zero-tag-region-p addr))
	       (integerp (seq-reachable-rec mem eip (code-size)))
	       (addr-p word))
	  (equal (static-inv-before (store-word-le mem addr word) eip)
		 (static-inv-before mem eip)))
 :hints (("Goal" :do-not-induct t)))

(defthm insn-sb-ok-local-same-after-data-store
 (implies (and (mem-p mem)
	       (integerp eip)
	       (code-region-p eip)
	       (integerp addr)
	       (or (data-region-p addr) (zero-tag-region-p addr))
	       (addr-p word))
	  (equal (insn-sandbox-ok-local (store-word-le mem addr word) eip)
		 (insn-sandbox-ok-local mem eip)))
 :hints (("Goal" :in-theory (enable insn-len insn-sandbox-ok-local)
	  :do-not-induct t)))

(defthm insn-sb-ok-inv-same-after-data-store
 (implies (and (mem-p mem)
	       (integerp eip)
	       (code-region-p eip)
	       (integerp addr)
	       (or (data-region-p addr) (zero-tag-region-p addr))
	       (addr-p word))
	  (equal (insn-sandbox-ok-inv (store-word-le mem addr word) eip inv)
		 (insn-sandbox-ok-inv mem eip inv)))
 :hints (("Goal" :in-theory (enable insn-len insn-sandbox-ok-inv)
	  :do-not-induct t)))

(defthm insn-sb-ok-same-after-data-store
 (implies (and (mem-p mem)
	       (integerp eip)
	       (code-region-p eip)
	       (integerp addr)
	       (or (data-region-p addr) (zero-tag-region-p addr))
	       (addr-p word))
	  (equal (insn-sandbox-ok (store-word-le mem addr word) eip)
		 (insn-sandbox-ok mem eip)))
 :hints (("Goal" :in-theory (enable insn-sandbox-ok)
	  :do-not-induct t)))

(in-theory (disable check-static-inv))

(defthm ebx-code-safe-implies-seq-reach
 (implies (and (static-inv-ebx-code-safe inv)
	       (check-static-inv inv x86)
	       (all-aligned-reachable-rec (x86-state-mem x86) (code-size))
	       (not (zero-tag-region-p (x86-state-ebx x86)))
	       (x86-state-p x86))
	  (seq-reachable-rec (x86-state-mem x86) (x86-state-ebx x86)
			     (code-size)))
 :hints (("Goal" :in-theory (enable check-static-inv))))

(defthm mem-sb-ok-after-data-store
 (implies (and (mem-p mem)
	       (natp k)
	       (integerp addr)
	       (or (data-region-p addr) (zero-tag-region-p addr))
	       (addr-p word))
	  (equal (mem-sandbox-ok-rec (store-word-le mem addr word) k)
		 (mem-sandbox-ok-rec mem k))))

(defthm all-align-reach-same-after-data-store
 (implies (and (mem-p mem)
	       (addr-p addr)
	       (or (data-region-p addr) (zero-tag-region-p addr))
	       (addr-p word))
	  (equal (all-aligned-reachable-rec (store-word-le mem addr word) k)
		 (all-aligned-reachable-rec mem k))))

(defthm logand-w-data-mask-is-data-or-zt-region
 (implies (and (natp addr)
	       (not (data-region-p (logand addr #x200000ff))))
	  (zero-tag-region-p (logand addr #x200000ff)))
 :hints (("Goal" :in-theory (enable data-region-p data-start data-size
				    data-end))))

(defthm safe-loop-implies-stuck
  (implies (and (x86-state-p x86)
		(safe-loop x86))
	   (equal (step-x86 x86) x86))
 :hints (("Goal" :in-theory (enable step-x86))))


;;; Properties of reachability

;; The next few lemmas give more properties about the reachability of
;; instructions and their appearance in the sequential
;; ordering. Specifically, they prove that 'kth-insn-from' and
;; 'eip-to-k' are inverse operations (for instructions that are in
;; fact reachable), and that adding one in the sequential ordering
;; corresponds to advancing 'eip' to the next instruction.

(in-theory (disable reset-static-inv))

(defthm nil-not-reachable
 (not (seq-reachable-rec mem nil k)))

(defthm kth-insn-of-seq-reach-rec-is-insn
 (implies (seq-reachable-rec mem eip k)
	  (equal (kth-insn-from mem (code-start) (seq-reachable-rec mem eip k))
		  eip)))

(defthm elim-kth-insn-from-given-seq-reach-rec
 (implies (and (seq-reachable-rec mem eip (code-size))
	       (equal k (seq-reachable-rec mem eip (code-size))))
	  (equal (kth-insn-from mem (code-start) k) eip)))

(defthm if-seq-reach-in-k-then-bound-by-kth-insn
 (implies (and (mem-p mem)
	       (natp k)
	       (natp eip)
	       (kth-insn-from mem (code-start) k)
	       (seq-reachable-rec mem eip k))
	  (<= eip (kth-insn-from mem (code-start) k))))

(defthm insn-k-plus-one-not-reachable-in-k-given-kth-exists
 (implies (and (mem-p mem)
	       (kth-insn-from mem (code-start) k)
	       (natp k))
	  (not (seq-reachable-rec
		mem (kth-insn-from mem (code-start) (+ 1 k)) k)))
 :hints (("Goal" :cases ((<= (kth-insn-from mem (code-start) (+ k 1))
			     (kth-insn-from mem (code-start) k))))))

(defthm insn-k-plus-one-not-reachable-in-k
 (implies (and (mem-p mem)
	       (natp k))
	  (not (seq-reachable-rec
		mem (kth-insn-from mem (code-start) (+ 1 k)) k)))
 :hints (("Goal" :cases ((kth-insn-from mem (code-start) k)))
	 ("Subgoal 1" :in-theory (disable kth-insn-from))))
				       
(defthmd seq-reach-rec-same-with-inc-bound
 (implies (and (seq-reachable-rec mem eip k)
	       (mem-p mem)
	       (natp k))
	  (equal (seq-reachable-rec mem eip (+ k 1))
		 (seq-reachable-rec mem eip k)))
 :hints (("Goal" :in-theory (disable kth-insn-from))))

(defthmd seq-reach-rec-same-with-inc-bound-2
 (implies (and (seq-reachable-rec mem eip (+ -1 k))
	       (mem-p mem)
	       (natp (+ -1 k)))
	  (equal (seq-reachable-rec mem eip k)
		 (seq-reachable-rec mem eip (+ -1 k))))
 :hints (("Goal" :in-theory (disable kth-insn-from)
	  :use (:instance seq-reach-rec-same-with-inc-bound (k (+ -1 k))))))

(defthmd seq-reach-rec-same-with-added-bound
 (implies (and (seq-reachable-rec mem eip k)
	       (mem-p mem)
	       (natp j)
	       (natp k))
	  (equal (seq-reachable-rec mem eip (+ j k))
		 (seq-reachable-rec mem eip k)))
 :hints (("Goal" :induct (natural-induction j)
	  :in-theory (enable seq-reach-rec-same-with-inc-bound-2))))

(defthmd rewrite-seq-reach-rec-w-size-bound-to-lower-bound
 (implies (and (seq-reachable-rec mem eip k)
	       (natp k)
	       (< k (code-size))
	       (mem-p mem))
	  (equal (seq-reachable-rec mem eip (code-size))
		 (seq-reachable-rec mem eip k)))
 :hints (("Goal" :use (:instance seq-reach-rec-same-with-added-bound
				 (j (- (code-size) k))))))

(defthm seq-reach-rec-of-kth-insn-is-k-bound-code-size-given-reach-k
 (implies (and (mem-p mem)
	       (natp k)
	       (seq-reachable-rec mem (kth-insn-from mem (code-start) k) k)
	       (< k (code-size))
	       (kth-insn-from mem (code-start) k))
	  (equal (seq-reachable-rec mem (kth-insn-from mem (code-start) k) 
				    (code-size))
		 k))
 :hints (("Goal" :do-not-induct t :use
	  (:instance rewrite-seq-reach-rec-w-size-bound-to-lower-bound
		     (eip (kth-insn-from mem (code-start) k))))))

(defthm seq-reach-rec-of-kth-insn-is-k-bound-code-size-given-code-size
 (implies (and (mem-p mem)
	       (natp k)
	       (seq-reachable-rec mem (kth-insn-from mem (code-start) k) 
				  (code-size)))
	  (equal (seq-reachable-rec mem (kth-insn-from mem (code-start) k) 
				    (code-size))
		 k))
 :hints (("Goal"
	  :cases ((and (< k (code-size))
		       (kth-insn-from mem (code-start) k)
		       (seq-reachable-rec mem
					  (kth-insn-from mem (code-start) k)
					  k))))
	 ("Subgoal 1"
	  :in-theory
	  (union-theories
	   (theory 'minimal-theory)
	   '(seq-reach-rec-of-kth-insn-is-k-bound-code-size-given-reach-k)))))

(encapsulate
 ()
 (local
  (defthmd lemma-1
    (implies (and (mem-p mem)
		  (natp k)
		  (seq-reachable-rec
		   mem
		   (+ (kth-insn-from mem (code-start) k)
		      (insn-len mem (kth-insn-from mem (code-start) k)))
		   (code-size))
		  (seq-reachable-rec mem (kth-insn-from mem (code-start) k)
				     (code-size)))
	     (equal (seq-reachable-rec
		     mem (+ (kth-insn-from mem (code-start) k)
			    (insn-len mem (kth-insn-from mem (code-start) k)))
		     (code-size))
		    (seq-reachable-rec
		     mem (kth-insn-from mem (code-start) (+ 1 k))
		     (code-size))))))
 (local
  (defthm lemma-2
    (implies (and (mem-p mem)
		  (natp k)
		  (seq-reachable-rec
		   mem (kth-insn-from mem (code-start) (+ 1 k))
		   (code-size))
		  (seq-reachable-rec
		   mem
		   (+ (kth-insn-from mem (code-start) k)
		      (insn-len mem (kth-insn-from mem (code-start) k)))
		   (code-size))
		  (seq-reachable-rec mem (kth-insn-from mem (code-start) k)
				     (code-size)))
	     (equal (seq-reachable-rec
		     mem (kth-insn-from mem (code-start) (+ 1 k))
		     (code-size))
		    (+ 1 (seq-reachable-rec
			  mem (kth-insn-from mem (code-start) k)
			  (code-size)))))
    :hints (("Goal" :in-theory (disable kth-insn-from)
	     :do-not-induct t))))
  (defthm seq-reach-rec-of-next-as-kth-is-one-plus
   (implies (and (mem-p mem)
		 (natp k)
		 (seq-reachable-rec
		  mem
		  (+ (kth-insn-from mem (code-start) k)
		     (insn-len mem (kth-insn-from mem (code-start) k)))
		  (code-size))
		 (seq-reachable-rec mem (kth-insn-from mem (code-start) k)
				    (code-size)))
	    (equal (seq-reachable-rec
		    mem (+ (kth-insn-from mem (code-start) k)
			   (insn-len mem (kth-insn-from mem (code-start) k)))
		    (code-size))
		   (+ 1 (seq-reachable-rec
			 mem (kth-insn-from mem (code-start) k) (code-size)))))
   :hints (("Goal" :cases ((seq-reachable-rec
			    mem (kth-insn-from mem (code-start) (+ 1 k))
			    (code-size))))
	   ("Subgoal 1" :in-theory (union-theories
				    (theory 'minimal-theory)
				    '(lemma-1 lemma-2)))))
 )

(defthm seq-reach-rec-of-next-is-one-plus
 (implies
  (and (mem-p mem)
       (seq-reachable-rec mem eip (code-size))
       (seq-reachable-rec
	mem
	(+ (kth-insn-from mem (code-start)
			  (seq-reachable-rec mem eip (code-size)))
	   (insn-len
	    mem
	    (kth-insn-from mem (code-start)
			   (seq-reachable-rec mem eip (code-size)))))
	(code-size))
       (seq-reachable-rec
	mem
	(kth-insn-from mem (code-start)
		       (seq-reachable-rec mem eip (code-size))) (code-size)))
  (equal
   (seq-reachable-rec
    mem
    (+ (kth-insn-from mem (code-start)
		      (seq-reachable-rec mem eip (code-size)))
       (insn-len
	mem
	(kth-insn-from mem (code-start)
		       (seq-reachable-rec mem eip (code-size))))) (code-size))
   (+ 1
      (seq-reachable-rec
       mem
       (kth-insn-from mem (code-start)
		      (seq-reachable-rec mem eip (code-size))) (code-size)))))
 :hints (("Goal" :do-not-induct t :in-theory
	  (disable elim-kth-insn-from-given-seq-reach-rec
		   kth-insn-of-seq-reach-rec-is-insn)))
 :rule-classes
 ((:rewrite
   :corollary
   (implies (and (mem-p mem)
		 (seq-reachable-rec mem (+ eip (insn-len mem eip)) (code-size))
		 (seq-reachable-rec mem eip (code-size)))
	    (equal (seq-reachable-rec mem (+ eip (insn-len mem eip))
				      (code-size))
		   (+ 1 (seq-reachable-rec mem eip (code-size)))))
   :hints (("Goal" :do-not-induct t
	    :cases ((equal eip (kth-insn-from
				mem (code-start)
				(seq-reachable-rec mem eip (code-size))))))))))

(defthm simplify-one-minus-seq-reach-of-next
 (implies (and (seq-reachable-rec mem eip (code-size))
	       (seq-reachable-rec mem (+ eip (insn-len mem eip))
				  (code-size))	       
	       (mem-p mem)
	       )
	  (equal (+ -1 (seq-reachable-rec mem (+ eip (insn-len mem eip))
					   (code-size)))
		 (seq-reachable-rec mem eip (code-size)))))

;;; Preservation lemmas

;; We now get to the most central facts of the proof: that each of the
;; elements of the safety invariant is preserved by execution steps.

(defthm reset-passes-ebp-data-safe
 (equal (static-inv-ebp-data-safe (reset-static-inv inv))
	(static-inv-ebp-data-safe inv))
 :hints (("Goal" :in-theory (enable reset-static-inv))))

(defthm ebx-data-safe-is-transient
 (not (static-inv-ebx-data-safe (reset-static-inv inv)))
 :hints (("Goal" :in-theory (enable reset-static-inv))))

(defthm ebx-code-safe-is-transient
 (not (static-inv-ebx-code-safe (reset-static-inv inv)))
 :hints (("Goal" :in-theory (enable reset-static-inv))))

(in-theory (disable zero-tag-region-p))

(defthm logand-w-code-mask-is-zt-or-code
  (implies (and (natp addr)
		(not (code-region-p (logand #x100000f0 addr))))
	   (zero-tag-region-p (logand #x100000f0 addr)))
  :hints (("Goal" :do-not-induct t
           :use (:instance logand-w-code-mask-is-code-or-zt))))

(defthm check-static-inv-ebx-code-safe-maybe
 (implies (and (not (static-inv-ebx-data-safe inv))
	       (not (static-inv-ebp-data-safe inv))
	       (natp addr))
	  (check-static-inv
	   inv
	   (x86-state eip eflags eax (logand #x100000f0 addr) ecx edx
		      esi edi ebp esp mem)))
 :hints (("Goal" :in-theory (enable check-static-inv))))

(in-theory (disable chunk-aligned-p))

;; Preservation lemma #1: if the current instruction is sequentially
;; reachable, so is the next one.
;;
;; Instructions 233 and 235- are short and long-displacement
;; unconditional direct jumps.
(defthm seq-reach-preservation
 (implies (and (step-x86 x86-state)
	       (x86-state-p x86-state)
	       (not (code-guard-p (x86-state-eip x86-state)))
	       (not (code-guard-p (x86-state-eip (step-x86 x86-state))))
	       (not (zero-tag-region-p (x86-state-eip (step-x86 x86-state))))
	       (mem-sandbox-ok-rec (x86-state-mem x86-state) (code-size))
	       (all-aligned-reachable-rec (x86-state-mem x86-state)
					  (code-size))
	       (seq-reachable-rec (x86-state-mem x86-state)
				  (x86-state-eip x86-state)
				  (code-size))
	       (insn-sandbox-ok (x86-state-mem x86-state)
				(x86-state-eip x86-state))
	       (check-static-inv (static-inv-before (x86-state-mem x86-state)
						    (x86-state-eip x86-state))
				 x86-state))
	  (seq-reachable-rec (x86-state-mem (step-x86 x86-state))
			     (x86-state-eip (step-x86 x86-state))
			     (code-size)))
 :hints (("Goal" :cases ((regular-step x86-state)))
	 ("Subgoal 1" :cases ((equal (mem-fetch (x86-state-mem x86-state)
						(x86-state-eip x86-state))
				     233)
			      (equal (mem-fetch (x86-state-mem x86-state)
						(x86-state-eip x86-state))
				     235)))
	 ("Subgoal 1.3" :in-theory (enable step-x86 regular-step
					   insn-sandbox-ok
					   insn-sandbox-ok-inv
					   insn-sandbox-ok-local))
	 ("Subgoal 1.2" :in-theory (enable step-x86 regular-step
					   insn-sandbox-ok
					   insn-sandbox-ok-inv
					   insn-sandbox-ok-local))
	 ("Subgoal 1.1" :in-theory (enable step-x86 regular-step
					   insn-sandbox-ok
					   insn-sandbox-ok-inv
					   insn-sandbox-ok-local))))
(defthm seq-reach-of-reg-step
 (implies (and (step-x86 x86-state)
	       (x86-state-p x86-state)
	       (not (code-guard-p (x86-state-eip x86-state)))
	       (not (code-guard-p (x86-state-eip (step-x86 x86-state))))
	       (not (zero-tag-region-p (x86-state-eip (step-x86 x86-state))))
	       (regular-step x86-state)
	       (not (safe-loop x86-state))
	       (mem-sandbox-ok-rec (x86-state-mem x86-state) (code-size))
	       (all-aligned-reachable-rec (x86-state-mem x86-state)
					  (code-size))
	       (seq-reachable-rec (x86-state-mem x86-state)
				  (x86-state-eip x86-state)
				  (code-size))
	       (insn-sandbox-ok (x86-state-mem x86-state)
				(x86-state-eip x86-state))
	       (check-static-inv (static-inv-before (x86-state-mem x86-state)
						    (x86-state-eip x86-state))
				 x86-state))
	  (seq-reachable-rec (x86-state-mem (regular-step x86-state))
			     (x86-state-eip (regular-step x86-state))
			     (code-size)))
 :hints (("Goal" :use (:instance seq-reach-preservation)
	  :in-theory (e/d (step-x86)
			  (seq-reach-preservation)))))

;; Preservation lemma #2: if the code region passes the sandboxing
;; checks before an instruction, it still does afterwards.
(defthm mem-sb-preservation
 (implies (and (step-x86 x86-state)
	       (safety-invariant x86-state))
	  (mem-sandbox-ok-rec (x86-state-mem (step-x86 x86-state))
			      (code-size)))
 :hints (("Goal" :in-theory (enable step-x86 regular-step))))

;; Preservation lemma #3: if all of the 16-byte aligned instructions
;; in the code region are reachable before an instruction, they still
;; are after it.
(defthm all-align-reach-preservation
 (implies (and (step-x86 x86-state)
	       (safety-invariant x86-state))
	  (all-aligned-reachable-rec (x86-state-mem (step-x86 x86-state))
				     (code-size)))
 :hints (("Goal" :in-theory (enable step-x86 regular-step))))

(defthm code-region-is-not-guard
 (implies (code-region-p addr)
	  (not (code-guard-p addr)))
 :hints (("Goal" :in-theory (enable code-region-p code-guard-p))))

(defthm check-ebx-code-safe-implies-code-region-p
 (implies (and (static-inv-ebx-code-safe inv)
	       (check-static-inv
		inv
		(x86-state eip eflags eax ebx ecx edx
			   esi edi ebp esp mem))
	       (not (zero-tag-region-p ebx)))
	  (code-region-p ebx))
 :hints (("Goal" :in-theory (enable check-static-inv))))

;; Preservation lemma #4: if the program counter is in the code region
;; before an instruction, it still is afterwards.
(defthm code-region-p-preservation
 (implies (and (step-x86 x86-state)
	       (x86-state-p x86-state)
	       (mem-sandbox-ok-rec (x86-state-mem x86-state)
				   (code-size))
	       (all-aligned-reachable-rec (x86-state-mem x86-state)
					  (code-size))
	       (code-region-p (x86-state-eip x86-state))
	       (seq-reachable-rec (x86-state-mem x86-state)
				  (x86-state-eip x86-state)
				  (code-size))
	       (insn-sandbox-ok (x86-state-mem x86-state)
				(x86-state-eip x86-state))
	       (check-static-inv (static-inv-before (x86-state-mem x86-state)
						    (x86-state-eip x86-state))
				 x86-state)
	       (not (zero-tag-region-p (x86-state-eip (step-x86 x86-state))))
	       (not (code-guard-p (x86-state-eip (step-x86 x86-state)))))
	  (code-region-p (x86-state-eip (step-x86 x86-state))))
 :hints (("Goal" :in-theory (enable step-x86 regular-step insn-sandbox-ok
				    insn-sandbox-ok-local
				    insn-sandbox-ok-inv))))

;; The expression on the left-hand-side of the following equality
;; often results from the expansion of the function
;; 'static-inv-before'.
(defthm simplify-minus-one-plus-seq-reach-in-static-inv
 (implies (and (mem-p mem)
	       (addr-p eip)
	       (seq-reachable-rec mem eip (code-size))
	       (seq-reachable-rec mem (+ eip (insn-len mem eip)) (code-size)))
	  (equal
	   (static-inv-after-kth-insn
	    mem
	    (+ -1
	       (seq-reachable-rec mem (+ eip (insn-len mem eip)) (code-size))))
	   (static-inv-after-kth-insn
	    mem
	    (eip-to-k mem eip)))))

;; ACL2 does not by default expand the definitions of recursive
;; functions like static-inv-after-kth-insn ('siaki' for
;; short). However, it is often necessary to unroll it by one level to
;; expose the calculation of the static invariant for the current
;; instruction (static-inv-trans); when enabled, this rule performs
;; that expansion.
(defthmd expand-siaki-one-step
 (implies (and (mem-p mem)
	       (addr-p eip)
	       (seq-reachable-rec mem eip (code-size)))
	  (equal (static-inv-after-kth-insn
		   mem (seq-reachable-rec mem eip (code-size)))
		 (static-inv-trans
		  mem eip 
		  (reset-static-inv
		   (static-inv-after-kth-insn
		    mem
		    (- (seq-reachable-rec mem eip (code-size)) 1)))))))

(defthm logand-with-jump-mask-is-aligned
 (implies (natp eip)
	  (chunk-aligned-p (logand #x100000f0 eip)))
 :hints (("Goal" :in-theory (enable chunk-aligned-p chunk-size-bits))))

;; The most complicated of the preservation lemmas concerns the
;; meaning of the per-instruction static invariants. To make the proof
;; more manageable for ACL2, we split it up into threesmaller cases
;; for the three aspects of the invariant, but it is still relatively
;; time consuming. While most of the lemmas in the proof can be
;; verified in less than a second, this one requires 2-5 minutes
;; depending on the speed of machine used, more than any other.
(defthm inv-preservation-ebx-code
 (implies (and (step-x86 x86-state)
	       (regular-step x86-state)
	       (not (code-guard-p (x86-state-eip x86-state)))
	       (not (zero-tag-region-p (x86-state-eip x86-state)))
	       (seq-reachable-rec (x86-state-mem (step-x86 x86-state))
				  (x86-state-eip (step-x86 x86-state))
				  (code-size))
	       (safety-invariant x86-state))
	  (check-ebx-code-inv (static-inv-before
			     (x86-state-mem (step-x86 x86-state))
			     (x86-state-eip (step-x86 x86-state)))
			    (step-x86 x86-state)))
 :hints (("Goal" :in-theory
	  (enable regular-step step-x86 check-static-inv
		  static-inv-trans
		  expand-siaki-one-step
		  insn-sandbox-ok insn-sandbox-ok-local
		  insn-sandbox-ok-inv))))


(defthm logand-w-data-mask-is-data-or-zt-region-c2
 (implies (and (natp addr)
               (not (data-region-p (logand addr 536871167))))
          (zero-tag-region-p (logand addr 536871167)))
 :hints (("Goal" :by logand-w-data-mask-is-data-or-zt-region))
 :rule-classes
 ((:rewrite
   :corollary
   (implies (and (natp addr)
                 (not (data-region-p (logand addr #x200000ff))))
            (zero-tag-region-p (logand #x200000ff addr)))
   :hints (("Goal" :in-theory (union-theories
                               (theory 'minimal-theory)
                               '(commutativity-of-logand)))))))

(defthm inv-preservation-ebx-data
 (implies (and (step-x86 x86-state)
	       (regular-step x86-state)
	       (not (code-guard-p (x86-state-eip x86-state)))
	       (not (zero-tag-region-p (x86-state-eip x86-state)))
	       (seq-reachable-rec (x86-state-mem (step-x86 x86-state))
				  (x86-state-eip (step-x86 x86-state))
				  (code-size))
	       (safety-invariant x86-state))
	  (check-ebx-data-inv (static-inv-before
			     (x86-state-mem (step-x86 x86-state))
			     (x86-state-eip (step-x86 x86-state)))
			    (step-x86 x86-state)))
 :hints (("Goal" :in-theory
	  (enable regular-step step-x86 check-static-inv
		  static-inv-trans
		  expand-siaki-one-step
		  insn-sandbox-ok insn-sandbox-ok-local
		  insn-sandbox-ok-inv))))

(defthm inv-preservation-ebp
 (implies (and (step-x86 x86-state)
	       (regular-step x86-state)
	       (not (code-guard-p (x86-state-eip x86-state)))
	       (not (zero-tag-region-p (x86-state-eip x86-state)))
	       (seq-reachable-rec (x86-state-mem (step-x86 x86-state))
				  (x86-state-eip (step-x86 x86-state))
				  (code-size))
	       (safety-invariant x86-state))
	  (check-ebp-inv (static-inv-before
			  (x86-state-mem (step-x86 x86-state))
			  (x86-state-eip (step-x86 x86-state)))
			 (step-x86 x86-state)))
 :hints (("Goal" :in-theory
	  (enable regular-step step-x86 check-static-inv
		  static-inv-trans
		  expand-siaki-one-step
		  insn-sandbox-ok insn-sandbox-ok-local
		  insn-sandbox-ok-inv))))

(defthmd inv-preservation-with-srr-hyp
 (implies (and (step-x86 x86-state)
	       (regular-step x86-state)
	       (not (code-guard-p (x86-state-eip x86-state)))
	       (not (zero-tag-region-p (x86-state-eip x86-state)))
	       (seq-reachable-rec (x86-state-mem (step-x86 x86-state))
				  (x86-state-eip (step-x86 x86-state))
				  (code-size))
	       (safety-invariant x86-state))
	  (check-static-inv (static-inv-before
			     (x86-state-mem (step-x86 x86-state))
			     (x86-state-eip (step-x86 x86-state)))
			    (step-x86 x86-state)))
 :hints (("Goal" :in-theory
	  (union-theories
	   (theory 'minimal-theory)
	   '(inv-preservation-ebx-data
	     inv-preservation-ebx-code
	     inv-preservation-ebp
	     check-static-inv)))))

(defthm in-code-implies-not-zt
 (implies (code-region-p eip)
	  (not (zero-tag-region-p eip)))
 :hints (("Goal" :in-theory (enable code-region-p zero-tag-region-p
				    code-start code-end))))

;; Preservation lemma #5: if the state before an instruction satisfies
;; the static invariant at that instruction, the state after the
;; instruction satisfies the invariant at the next instruction.
(defthm inv-preservation
 (implies (and (step-x86 x86-state)
	       (regular-step x86-state)
	       (code-region-p (x86-state-eip (step-x86 x86-state)))
	       (not (code-guard-p (x86-state-eip x86-state)))
	       (not (zero-tag-region-p (x86-state-eip x86-state)))
	       (safety-invariant x86-state))
	  (check-static-inv (static-inv-before
			     (x86-state-mem (step-x86 x86-state))
			     (x86-state-eip (step-x86 x86-state)))
			    (step-x86 x86-state)))
 :hints (("Goal" :use ((:instance inv-preservation-with-srr-hyp)
		       (:instance seq-reach-preservation)))))

(in-theory (disable check-static-inv static-inv-before))

(defthmd inv-preservation-2
 (implies
   (and (step-x86 x86-state)
        (x86-state-p x86-state)
        (mem-sandbox-ok-rec (x86-state-mem x86-state)
                            (code-size))
        (all-aligned-reachable-rec (x86-state-mem x86-state)
                                   (code-size))
        (code-region-p (x86-state-eip x86-state))
        (seq-reachable-rec (x86-state-mem x86-state)
                           (x86-state-eip x86-state)
                           (code-size))
        (insn-sandbox-ok (x86-state-mem x86-state)
                         (x86-state-eip x86-state))
        (check-static-inv (static-inv-before (x86-state-mem x86-state)
                                             (x86-state-eip x86-state))
                          x86-state)
	(not (zero-tag-region-p (x86-state-eip x86-state)))
	(code-region-p (x86-state-eip (step-x86 x86-state)))
        (not (code-guard-p (x86-state-eip (step-x86 x86-state)))))
   (check-static-inv (static-inv-before (x86-state-mem (step-x86 x86-state))
                                        (x86-state-eip (step-x86 x86-state)))
                     (step-x86 x86-state)))
 :hints (("Goal" :cases ((not (regular-step x86-state))))
	 ("Subgoal 1" :in-theory (enable step-x86 insn-sandbox-ok
					 insn-sandbox-ok-local
					 insn-sandbox-ok-inv))))

;; Preservation lemma #6: if the current instruction is legal
;; according to the sandboxing policy, the next one will be too. Since
;; mem-sandbox-ok guarantees that every sequentially reachable
;; instruction is legal, this is actually a simple corollary of
;; preservation lemmas 1 and 2.
(defthmd insn-sb-ok-preservation
 (implies (and (x86-state-p x86-state)
	       (step-x86 x86-state)
	       (insn-sandbox-ok (x86-state-mem x86-state)
				(x86-state-eip x86-state))
	       (mem-sandbox-ok-rec (x86-state-mem x86-state)
				   (code-size))
	       (all-aligned-reachable-rec (x86-state-mem x86-state)
					  (code-size))
	       (seq-reachable-rec (x86-state-mem x86-state)
				  (x86-state-eip x86-state)
				  (code-size))
	       (check-static-inv (static-inv-before (x86-state-mem x86-state)
						    (x86-state-eip x86-state))
				 x86-state)
	       (not (zero-tag-region-p (x86-state-eip (step-x86 x86-state))))
	       (not (code-guard-p (x86-state-eip (step-x86 x86-state)))))
	  (insn-sandbox-ok (x86-state-mem (step-x86 x86-state))
			   (x86-state-eip (step-x86 x86-state))))
 :hints (("Goal" :cases
	  ((not (and (seq-reachable-rec (x86-state-mem (step-x86 x86-state))
					(x86-state-eip (step-x86 x86-state))
					(code-size))
		     (mem-sandbox-ok-rec (x86-state-mem (step-x86 x86-state))
					 (code-size))))))
	 ("Subgoal 2"
	  :in-theory (union-theories (theory 'minimal-theory)
				     '(seq-reach-in-sb-is-ok)))
	 ("Subgoal 1" :use ((:instance mem-sb-preservation)
			    (:instance seq-reach-preservation)))))

;;; Final safety induction

;; Once all of the individual preservation lemmas have been proved,
;; the remainder of the proof is relatively easy.

;; The statement of the preservation of the complete safety invariant
;; is quite simple:
(defthm safety-preservation
 (implies (and (step-x86 x86-state)
	       (safety-invariant x86-state))
	  (safety-invariant (step-x86 x86-state)))
 :hints (("Goal" :in-theory
	  (enable mem-sb-preservation code-region-p-preservation
		  inv-preservation-2 insn-sb-ok-preservation))))

(defthm if-sb-then-start-seq-reach
 (implies (and (mem-p mem)
	       (natp k)
	       (mem-sandbox-ok mem))
	  (seq-reachable-rec mem (code-start) k)))

(defthm mem-sb-ok-means-first-insn-sb-ok
 (implies (and (mem-p mem)
	       (mem-sandbox-ok-rec mem k))
	  (insn-sandbox-ok mem (code-start))))

(defthm srr-of-start-is-zero
 (implies (and (mem-p mem)
	       (natp k))
	  (equal (seq-reachable-rec mem (code-start) k)
		 0))
 :hints (("Goal" :in-theory (enable code-size))))
 
(defthm first-static-inv-is-trivial
 (implies (x86-state-p x86)
	  (equal (static-inv-before (x86-state-mem x86) (code-start))
		 (base-static-inv)))
 :hints (("Goal" :in-theory (enable static-inv-before))))

;; The base case for the safety induction: the initial state satisfies
;; the safety invariant.
(defthm start-state-inv
 (implies (is-start-state x86)
	  (safety-invariant x86))
 :hints (("Goal" :in-theory (enable x86-state-p check-static-inv))))

(in-theory (disable safety-invariant is-start-state))

;; The inductive case for the safety induction: if one state satisfies
;; the invariant, so does the next.
(defthm safety-induction
 (implies (safety-invariant x86-state)
	  (safety-invariant (step-x86 x86-state)))
 :hints (("Goal" :in-theory (enable safety-invariant))))

;; Safety in the sense of not making any unsafe steps follows at each
;; step from the safety invariant:
(defthm safety-inv-implies-safe-to-now
 (implies (safety-invariant x86)
	  (consp x86))
 :hints (("Goal" :in-theory (enable safety-invariant))))

;; The safety invariant holds at each step:
(defthm if-sb-then-safety-inv-after-k
  (implies (is-start-state x86)
	   (safety-invariant (step-for-k x86 k))))

;; The following statement is the final safety result for the
;; system. It states that whatever the initial configuration of the
;; registers and memory, if the code in memory passes the static
;; safety check, then execution of the code can continue for any
;; number K of instructions without violating the safety policy. The
;; ACL2 function consp is true for a state object (represented as a
;; list), but false for nill.
;;
;; The hypothesis '(data-region-p ebp)' is necessary because we have
;; defined the invariant on %ebp to hold at the beginning of execution
;; (see the definition of base-static-inv); this matches the choice
;; made by the real system. If the verification were defined to
;; require the sandboxed code to establish this fact, the hypothesis
;; could be removed.
(defthm safety
  (implies (and (mem-p mem)
		(mem-sandbox-ok mem)
		(data-region-p ebp)
		(addr-p eax) (addr-p ebx) (addr-p ecx) (addr-p edx)
		(addr-p esi) (addr-p edi) (addr-p ebp) (addr-p esp)
		(addr-p eflags))
	   (consp
	    (step-for-k
	     (x86-state (code-start) eflags eax ebx ecx edx
			esi edi ebp esp mem)
	     k)))
  :hints (("Goal" :do-not-induct t
	   :in-theory (enable is-start-state x86-state-p))))
